!
! Module: fits_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare aa ptrs => null()!
! 
! $Id: fits_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fits_utils

  use dal

!  use sscmisc

  implicit none 

  private 

  public readColData
  public getRealKwd
  public getIntKwd
  public getStringKwd
  public loadFitsArray
  public splitSetTabName
!  public saveOutImage
  public copyAttribute
  public copyVitalArrayKwds
  public columnSetIsIdentical

  interface readColData
    module procedure readColData_real32_wn
    module procedure readColData_real64_wn
    module procedure readColData_int8_wn
    module procedure readColData_int16_wn
    module procedure readColData_int32_wn
    module procedure readColData_real32
    module procedure readColData_real64
    module procedure readColData_int8
    module procedure readColData_int16
    module procedure readColData_int32
    module procedure readColData2dReal32
    module procedure readColData2dReal64
  end interface

  interface loadFitsArray
    module procedure loadFitsArray2dSingleName, loadFitsArray2dSingleArray
    module procedure loadFitsArray2dBoolean, loadFitsArray2dInt32Array
  end interface

  interface copyAttribute
    module procedure copyAttribute_arrayToArray
    module procedure copyAttribute_arrayToSet
    module procedure copyAttribute_setToSet
    module procedure copyAttributeTabToTab
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine readColData2dReal32(tab, colName, colPtr)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 pointer colPtr. NOTE that this might give
  ! funny results if the column values are REAL64 for example and out of
  ! the REAL32 range.
  !
  ! Note that colPtr() should be deallocated after use by the calling routine.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), pointer    :: colPtr(:,:)

    type(ColumnT) :: col
    integer, pointer :: dimVector(:)
    integer(int8), pointer ::&
      colDataInt8(:), vectorColDataInt8(:,:)
    integer(int16), pointer ::&
      colDataInt16(:), vectorColDataInt16(:,:)
    integer(int32), pointer ::&
      colDataInt32(:), vectorColDataInt32(:,:)
    real(single), pointer ::&
      colDataReal32(:), vectorColDataReal32(:,:)
    real(double), pointer ::&
      colDataReal64(:), vectorColDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    dimVector => dimensions(col)
    if (size(dimVector) > 1) then
call error('badColDimension','ssclib:')
    end if

    if (size(dimVector) == 0) then
      allocate(colPtr(1, numberOfRows(tab)))

      select case(dataType(col))
      case (INTEGER8)
        colDataInt8 => int8Data(col)
        colPtr(1, :) = real(colDataInt8, single)
      case (INTEGER16)
        colDataInt16 => int16Data(col)
        colPtr(1, :) = real(colDataInt16, single)
      case (INTEGER32)
        colDataInt32 => int32Data(col)
        colPtr(1, :) = real(colDataInt32, single)
      case (REAL32)
        colDataReal32 => real32Data(col)
        colPtr(1, :) = colDataReal32
      case default ! this gives a DAL error if data type is none of the above.
        colDataReal64 => real64Data(col)
        colPtr(1, :) = real(colDataReal64, single)
      end select

    else
      allocate(colPtr(dimVector(0), numberOfRows(tab)))

      select case(dataType(col))
      case (INTEGER8)
        vectorColDataInt8 => int8VectorData(col)
        colPtr = real(vectorColDataInt8, single)
      case (INTEGER16)
        vectorColDataInt16 => int16VectorData(col)
        colPtr = real(vectorColDataInt16, single)
      case (INTEGER32)
        vectorColDataInt32 => int32VectorData(col)
        colPtr = real(vectorColDataInt32, single)
      case (REAL32)
        vectorColDataReal32 => real32VectorData(col)
        colPtr = vectorColDataReal32
      case default ! this gives a DAL error if data type is none of the above.
        vectorColDataReal64 => real64VectorData(col)
        colPtr = real(vectorColDataReal64, single)
      end select
    end if

    call release(col)

  end subroutine readColData2dReal32

  !---------------------------------------------------------------------
  subroutine readColData2dReal64(tab, colName, colPtr)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 pointer colPtr.
  !
  ! Note that colPtr() should be deallocated after use by the calling routine.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), pointer    :: colPtr(:,:)

    type(ColumnT) :: col
    integer, pointer :: dimVector(:)
    integer(int8), pointer ::&
      colDataInt8(:), vectorColDataInt8(:,:)
    integer(int16), pointer ::&
      colDataInt16(:), vectorColDataInt16(:,:)
    integer(int32), pointer ::&
      colDataInt32(:), vectorColDataInt32(:,:)
    real(single), pointer ::&
      colDataReal32(:), vectorColDataReal32(:,:)
    real(double), pointer ::&
      colDataReal64(:), vectorColDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    dimVector => dimensions(col)
    if (size(dimVector) > 1) then
call error('badColDimension','ssclib:')
    end if

    if (size(dimVector) == 0) then
      allocate(colPtr(1, numberOfRows(tab)))

      select case(dataType(col))
      case (INTEGER8)
        colDataInt8 => int8Data(col)
        colPtr(1, :) = real(colDataInt8, double)
      case (INTEGER16)
        colDataInt16 => int16Data(col)
        colPtr(1, :) = real(colDataInt16, double)
      case (INTEGER32)
        colDataInt32 => int32Data(col)
        colPtr(1, :) = real(colDataInt32, double)
      case (REAL32)
        colDataReal32 => real32Data(col)
        colPtr(1, :) = real(colDataReal32, double)
      case default ! this gives a DAL error if data type is none of the above.
        colDataReal64 => real64Data(col)
        colPtr(1, :) = colDataReal64
      end select

    else
      allocate(colPtr(dimVector(0), numberOfRows(tab)))

      select case(dataType(col))
      case (INTEGER8)
        vectorColDataInt8 => int8VectorData(col)
        colPtr = real(vectorColDataInt8, double)
      case (INTEGER16)
        vectorColDataInt16 => int16VectorData(col)
        colPtr = real(vectorColDataInt16, double)
      case (INTEGER32)
        vectorColDataInt32 => int32VectorData(col)
        colPtr = real(vectorColDataInt32, double)
      case (REAL32)
        vectorColDataReal32 => real32VectorData(col)
        colPtr = real(vectorColDataReal32, double)
      case default ! this gives a DAL error if data type is none of the above.
        vectorColDataReal64 => real64VectorData(col)
        colPtr = vectorColDataReal64
      end select
    end if

    call release(col)

  end subroutine readColData2dReal64

  !---------------------------------------------------------------------
  logical function columnSetIsIdentical(tableA, tableB)
  !---------------------------------------------------------------------
  ! Note this does not compare the column VALUES, just their occurrence,
  ! dimensions and data type.
  !
  ! NOTE that this function allocates some column and dimensions() pointers.
  ! These should of course be cleared by releasing the tables or even
  ! the datasets at some stage.

    implicit none

    type(TableT), intent(in) :: tableA, tableB

    integer :: numColsA, numColsB, colNum
    type(ColumnT) :: colA, colB
    character(255) :: colName
    integer, pointer :: dimsA(:), dimsB(:)

    !-------------------------------------------------------------------
    ! Start

    columnSetIsIdentical = .false. ! default

    numColsA = numberOfColumns(tableA)
    numColsB = numberOfColumns(tableB)
  if (numColsA .ne. numColsB) return

    do colNum = 1, numColsA
      colA = column(tableA, colNum - 1)
      colName = name(colA)
  if (.not. hasColumn(tableB, colName)) return
  if (dataType(colA) .ne. dataType(column(tableB, colName))) return
      colB = column(tableB, colName)
      dimsA => dimensions(colA)
      dimsB => dimensions(colB)
  if (size(dimsA) .ne. size(dimsB)) return
  if (any(dimsA .ne. dimsB)) return
    end do

    columnSetIsIdentical = .true.

  end function columnSetIsIdentical

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToArray(inArray, kwdName, outArray, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outArray, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToArray

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToSet(inArray, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(DataSetT), intent(in) :: outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToSet

  !---------------------------------------------------------------------
  subroutine copyAttribute_setToSet(inSet, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: inSet, outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inSet, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inSet, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_setToSet

  !---------------------------------------------------------------------
  subroutine copyAttributeTabToTab(inTab, kwdName, outTab, status)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: inTab, outTab
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inTab, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outTab, attribute(inTab, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttributeTabToTab

  !---------------------------------------------------------------------
  subroutine saveOutImage(newImage, imageName, newImageName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xStartArg, yStartArg

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    type(BlockT) :: inBlock, outBlock
!    type(DssT) ::&
!      my_dss
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) :: xStart, yStart, xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
!    inBlock = block(inArray)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(newImage, 1)
    dimVector(1) = size(newImage, 2)
    xFinis = xStart + size(newImage, 1) - 1
    yFinis = yStart + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
!    outBlock = block(outArray)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = 0.0
    arrayDataReal32(xStart-1:xFinis-1, yStart-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImage

  !---------------------------------------------------------------------
  subroutine copyVitalArrayKwds(inArray, outArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray

    integer :: status

    !-------------------------------------------------------------------
    ! Start

    call copyAttribute(inArray, 'SLCTEXPR', outArray, status)
    call copyAttribute(inArray, 'HDUCLASS', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS1', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS2', outArray, status)
    call copyAttribute(inArray, 'HDUVERS1', outArray, status)

    call copyAttribute(inArray, 'CTYPE1', outArray, status)
    call copyAttribute(inArray, 'CRPIX1', outArray, status)
    call copyAttribute(inArray, 'CRVAL1', outArray, status)
    call copyAttribute(inArray, 'CUNIT1', outArray, status)
    call copyAttribute(inArray, 'CDELT1', outArray, status)
    call copyAttribute(inArray, 'CTYPE2', outArray, status)
    call copyAttribute(inArray, 'CRPIX2', outArray, status)
    call copyAttribute(inArray, 'CRVAL2', outArray, status)
    call copyAttribute(inArray, 'CUNIT2', outArray, status)
    call copyAttribute(inArray, 'CDELT2', outArray, status)

    call copyAttribute(inArray, 'EXPOSURE', outArray, status)

  end subroutine copyVitalArrayKwds

  !---------------------------------------------------------------------
  subroutine readColData_real32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real64_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real64_wn

  !---------------------------------------------------------------------
  subroutine readColData_int8_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int8_wn

  !---------------------------------------------------------------------
  subroutine readColData_int16_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int16_wn

  !---------------------------------------------------------------------
  subroutine readColData_int32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real32

  !---------------------------------------------------------------------
  subroutine readColData_real64(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real64

  !---------------------------------------------------------------------
  subroutine readColData_int8(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int8

  !---------------------------------------------------------------------
  subroutine readColData_int16(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int16

  !---------------------------------------------------------------------
  subroutine readColData_int32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int32

  !---------------------------------------------------------------------
  subroutine splitSetTabName(setTabName, setName, tabName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in)  :: setTabName
    character(*), intent(out) :: setName, tabName

    integer(int16) :: posOfColon
    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start

    posOfColon = index(setTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.
      call warning('missingBlockName', "No colon+block name was &
        &found after the dataset name. I'm going to use the first block &
        &in the dataset. No cause for alarm - provided that this &
        &really is the block you want!")
      setName = setTabName
      set = dataSet(setName, READ)
      tabName = name(block(set, 0))
      call release(set)

    else
      setName = setTabName(:posOfColon - 1)
      tabName = setTabName(posOfColon + 1:)
    end if

  end subroutine splitSetTabName

  !---------------------------------------------------------------------
  subroutine getRealKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS real32 or 64 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    real(double),   intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0.0d0 ! default

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. REAL32) then
        value = real32Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. REAL64) then
          value = real64Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. REAL32) then
        value = real32Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. REAL64) then
          value = real64Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getRealKwd

  !---------------------------------------------------------------------
  subroutine getIntKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS int8, 16 or 32 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    integer(int32), intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0 ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER8) then
        value = int8Attribute(inDataSet, keyword)
      else if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER16) then
        value = int16Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. INTEGER32) then
          value = int32Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. INTEGER8) then
        value = int8Attribute(imageArray, keyword)
      else if (dataType(attribute(imageArray, keyword)) .eq. INTEGER16) then
        value = int16Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. INTEGER32) then
          value = int32Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getIntKwd

  !---------------------------------------------------------------------
  subroutine getStringKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS string attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!****Here it is assumed that a keyword of any type can be accessed as string?? Should check this.

    implicit none

    character*(*),  intent(in)            :: keyword
    character*(*),  intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = "??" ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      value = stringAttribute(inDataSet, keyword)
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      value = stringAttribute(imageArray, keyword)
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getStringKwd

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dInt32Array(inArray, image)
  !---------------------------------------------------------------------
  ! NOTE the pointer image() should be explicitly deallocated by the
  ! calling routine after use. The other pointers should become
  ! deallocated when the array (or dataset) is released.

    implicit none

    type(ArrayT), intent(in) :: inArray
    integer(int32), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
!    integer(int32), dimension(:,:,:), pointer :: cube
!    integer(int32), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions!!!, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

!    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = int(arrayDataInt8, int32)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = int(arrayDataInt16, int32)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = arrayDataInt32
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = nint(arrayDataReal32, int32)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = nint(arrayDataReal64, int32)
      end if

!    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!
!    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray2dInt32Array

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dSingleName(setName, fitsArray, blockNumber)
  !---------------------------------------------------------------------
  ! Remember to deallocate the pointer fitsArray after use!

    implicit none

    character(*), intent(in) :: setName
    integer, intent(in), optional :: blockNumber
    real(single), pointer :: fitsArray(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer :: localBlockNumber
    integer(int32), dimension(:), pointer ::&
      dimVector
!!!    real(single), dimension(:,:,:), pointer :: cube
!!!    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4)
!!!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start

    if (present(blockNumber)) then
      localBlockNumber = blockNumber
    else
      localBlockNumber = 0
    end if

    set = dataSet(setName, READ)
    inArray = array(set, localBlockNumber, READ)
    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(fitsArray(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        fitsArray = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        fitsArray = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        fitsArray = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        fitsArray = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        fitsArray = real(arrayDataReal64, single)
      end if

    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(cube, imageShape)
call error('notSupported', '')

    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(hyperCube, imageShape)
call error('notSupported', '')

    case default
call error('arrayRankTooBig', '')
    end select

call release(set)
!*** very important! Make sure these releases are in the subroutines when you get around to putting them in a library.

  end subroutine loadFitsArray2dSingleName

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dSingleArray(inArray, image)
  !---------------------------------------------------------------------
  ! NOTE the pointer image() should be explicitly deallocated by the
  ! calling routine after use. The other pointers should become
  ! deallocated when the array (or dataset) is released.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
!    integer(int32), dimension(:,:,:), pointer :: cube
!    integer(int32), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions!!!, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

!    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = real(arrayDataReal64, single)
      end if

!    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!
!    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray2dSingleArray

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dBoolean(setName, fitsArray, blockNumber)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: setName
    integer, intent(in), optional :: blockNumber
    logical, pointer :: fitsArray(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer :: localBlockNumber
    integer(int32), dimension(:), pointer ::&
      dimVector
!!!    real(single), dimension(:,:,:), pointer :: cube
!!!    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4)
!!!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean

    !-------------------------------------------------------------------
    ! Start

    if (present(blockNumber)) then
      localBlockNumber = blockNumber
    else
      localBlockNumber = 0
    end if

    set = dataSet(setName, READ)
    inArray = array(set, localBlockNumber, READ)
    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(fitsArray(dimVector(0), dimVector(1)))

      fitsArray = .false.
      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        where(arrayDataInt8 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        where(arrayDataInt16 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        where(arrayDataInt32 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        where(arrayDataReal32 > 0) fitsArray = .true.
!      else if (dataType(inArray) .eq. BOOLEAN) then
!        arrayDataBoolean => boolArray2Data(inArray)
!        fitsArray = arrayDataBoolean
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        where(arrayDataReal64 > 0) fitsArray = .true.
      end if

    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(cube, imageShape)
call error('notSupported', '')

    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(hyperCube, imageShape)
call error('notSupported', '')

    case default
call error('arrayRankTooBig', '')
    end select

call release(set)
!*** very important! Make sure these releases are in the subroutines when you get around to putting them in a library.

  end subroutine loadFitsArray2dBoolean

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fits_utils
